/*============================================================================
Copyright 2006 Thad Harroun, Kevin Yager

This file is part of "yanera 2.0".

"yanera 2.0" is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version.

"yanera 2.0" is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with "yanera 2.0"; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
=============================================================================*/
/**   \file yanera.h
*     \brief Global header file.
 *    \author Thad Harroun <tharroun@brocku.ca>
 * 
 *  This file is a global header file, that outlines the \c yanera_container
 *  struct, and everything in it. It should be included in every source file.
 */ 
/*===========================================================================*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <limits.h>
#include <stdarg.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>
#include <gsl/gsl_complex_math.h>
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_multifit_nlin.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_sf_erf.h>
#include <gsl/gsl_sf_gamma.h>
#include <gsl/gsl_siman.h>
#include <gsl/gsl_vector.h>
#ifdef HAVE_OOL
#include <ool/ool_conmin.h>
#include <ool/ool_tools_diff.h>
#endif
#include <libxml/xmlreader.h>
#include <libxml/tree.h>

#define _CRT_SECURE_NO_DEPRECATE

/*============================================================================*/
/** -----------------------------------------------
 *  \brief Enum definitions for fit constraints. 
 *
 * Enum definitions for fit constraints.
 */
enum fit_constraint_type {
  /** PLEASE FILL IN */
  CONSTRAIN_NONE,
  /** PLEASE FILL IN */
  CONSTRAIN_MAX,
  /** PLEASE FILL IN */
  CONSTRAIN_MIN,
  /** PLEASE FILL IN */
  CONSTRAIN_BOTH
};
/** -----------------------------------------------
 *  \brief Enum definitions for data weighting. 
 *
 * Enum definitions for data weighting. The wieghts are based on the error of
 * the data, \f$\sigma\f$, or as defined below.
 */
enum fit_weighting_type {
  /** \f$\sigma\f$ = 1               */
  WEIGHT_NONE,
  /** \f$\sigma = y\f$         */
  WEIGHT_R,
  /** \f$\sigma = \sqrt{y}\f$  */
  WEIGHT_SQRT,
  /** \f$\sigma = \textrm{data file}\f$       */
  WEIGHT_DATA
};
/** -----------------------------------------------
 *  \brief Enum definitions for keeping track of resolution types. 
 *
 * Enum definitions for keeping track of the types of data resolution.
 */
enum resolution_type {
  /** No instrumental resolution smearing. */
  RESOLUTION_NONE,
  /** \f$\delta q = \mathrm{const.}\f$ */
  RESOLUTION_CONSTANT,
  /** \f$\delta q/q = \mathrm{const.}\f$ */ 
  RESOLUTION_RELATIVE,
  /** Treat as constant, but calculate from an array of values */
  RESOLUTION_ARRAY_CONSTANT,
  /** Treat as relative, but calculate from an array of values */
  RESOLUTION_ARRAY_RELATIVE 
};
/** -----------------------------------------------
 * \brief Enum definitions for keeping track of model types. 
 *
 * Enum definitions for keeping track of model types.
 */
enum model_type {
  /** General flag is no type is specified */
  MODEL_NONE,
  /** Direct : Ordinary parratt layer model */
  MODEL_LAYER,
  /** Tiny slab : Component (box and gaussian) model */
  MODEL_COMPONENT,
  /** Tiny slab : Layer model broken into tiny slabs */
  MODEL_SLAB,
  /** Tiny slab : Function model */
  MODEL_FUNCTION,
  /** When linking a new XML model to the list, 
      specifiy XML with this flag */
  MODEL_XML,
  /** When linking a new complete model to the list, 
     specifiy that it should be linked to the complete list with this flag */
  MODEL_COMPLETE 
};
/** -----------------------------------------------
 * \brief Enum definitions for keeping track of layer types. 
 *
 * Each layer type is going to be one of these enum types. 
 * The types are pretty much self-explanatory.
 */
enum layer_type {
  LAYER_NONE,      /**< FILL IN */
  LAYER_LAYER,     /**< FILL IN */
  LAYER_BULK,      /**< FILL IN */
  LAYER_REPEAT,    /**< FILL IN */
  LAYER_SUBSTRATE, /**< FILL IN */
  LAYER_COMPONENT, /**< FILL IN */
  LAYER_BOX,       /**< FILL IN */
  LAYER_GAUSSIAN,  /**< FILL IN */
  LAYER_FUNCTION   /**< FILL IN */
};
/** ----------------------------------------------- 
 * \brief Generic yes/no boolean for convienience. 
 *
 * Generic yes/no boolean for convienience
 */
enum boolean {
  YES,
  NO,
  ERROR
};
/** -----------------------------------------------
 *  \brief Enum definition for keeping track of polarization experiment type
 *
 * Polarization is indicated by these definitions.
 */
enum polarized_type {
  /** Unpolarized condition */
  POLARIZED_NONE,
  /** Total SLD = Real nuclear SLD + Magnetic SLD */
  POLARIZED_UP,
  /** Total SLD = Real nuclear SLD - Magnetic SLD */
  POLARIZED_DOWN,
  /** Total SLD = Real nuclear SLD only, used to output as xml */
  POLARIZED_UNPOLARIZED
};
/** -----------------------------------------------
 *  \brief Enum definition for fitting type
 *
 *  Enum definition for fitting types
 */
 enum fit_method_type {
  CONJUGATE_FR,  /**< FILL IN */
  CONJUGATE_PR,  /**< FILL IN */
  VECTOR_BFGS,   /**< FILL IN */
  STEEP_DESC,    /**< FILL IN */
  NMSIMPLEX,     /**< FILL IN */
  SIMAN,         /**< FILL IN */
  LMSDR,         /**< FILL IN */
  LMDR,          /**< FILL IN */
  OOL_SPG,       /**< FILL IN */
  OOL_GENCAN,    /**< FILL IN */
  CLMSDR,        /**< FILL IN */
  CLMDR,         /**< FILL IN */
  LEVMAR         /**< FILL IN */
 };
/*============================================================================*/
/** ---------------------------------------------------------------------------
 *  \brief Linked list element holding data for evaluating function expressions.
 *
 *  This holds all types of tokens in the postfix expression. This includes: 
 *  real numbers (eg. 1.2, -3.45E-3), 
 *  or parameters (eg. 'a', 'b'), 
 *  or operators (eg. +, *), 
 *  or functions (eg. 'cos', 'exp'),
 *  or the ordinate (eg. 'x').
 *  \see yanera_postfix
 */
struct _postfix_queue_element {
  /**
	 *  Token type.  See \c operator_type in \c yanera_postfix.h
	 */
	unsigned short type;
  /**
	 *  If token is a real number, store it here. 
	 */
	double         numb;
	/**
	 *  Pointer to parameter index if token is a parameter  
	 */
	short          indx;
	/**
	 *  Parameters and functions are strings, but no function
	 *  is more than 3 characters, plus null termination.
	 */ 
	char           oper[4];
	/**
	 *  Link to next token in the list.
	 */ 
	struct _postfix_queue_element *next;
};
/**  See \c _postfix_queue documentation */
typedef struct _postfix_queue_element yanera_postfix_queue_element;
/** ---------------------------------------------------------------------------
 * \brief Linked list for evaluating functional expreessions.
 *
 * In order to evaluate a user supplied functional form of a layer, 
 * the expression must be converted to postfix notation. This queue holds 
 * the expresion, token by token, for rapid and easy evaluation

 */
struct _postfix_queue{
	unsigned short                n;     /**< Number of list elements */
	yanera_postfix_queue_element *front; /**< Linked list link to front */
	yanera_postfix_queue_element *rear;  /**< Linked list link to rear */
};
/**  See \c _postfix_queue documentation */
typedef struct _postfix_queue yanera_postfix;
/** ---------------------------------------------------------------------------
 * \brief This struct is for all layer and component types
 *
 *
 *  This struct is used for every layer type, including \<layer/\>, 
 *  \<component/\>, and \<function/\>. The struct carries different types 
 *  of information, but the actual numerical parameters are stored in the
 *  paramter array, \c yanera_parameters, not here.  Instead, the 
 *  \c yanera_layer struct stores the indexes to the parameter array.
 */
struct _layer {
  /**
   *  The XML layer_id identifies layers in the base model to be altered.
   */
  char               *layer_id;  
  /**
   *  The XML layer_idref refers to a \<layer/\> with 
   *  a unique layer_id in the base model.
   */
  char               *layer_idref;
  /**
   *  The XML component_id is used to identify a common component.
   */
  char               *component_id;
  /**
   *  The XML component_idref alows a \<layer/\> to identify with 
   *  a common component
   */
  char               *component_idref;
  /**
   *  General layer information : Layer type. See \c layer_type
   */
  enum layer_type     type;
  /**
   *  General layer information : Layer name. Character string.
   */
  char               *name;
  /**
   *  Function type layers info : The actual text of a layer equation.
   */
  char               *func;
  /**
   *  Function type layers info : Postfix expression of \c func. 
   *  See \c yanera_postfix for more information.
   */
  yanera_postfix     *func_postfix;
  /**
   *  Function type layers info :
   *  Array of function parameter names, eg. 'a,b,c,...'
   */
  char               *parm_names;
  /**
   *  Function type layers info : 
   *  Array of pointers to the parameter array, one for each 'a,b,c,...'
   *  that is found in \c parm_name 
   */
  short              *parm_values;
  /**
   *  Function type layers info :
   *  Number of function parameter names, eg. 'a,b,c,d'=4
   */
  short               number_of_parm;
  /**
   *  Parameter pointer : Layer thickness or box width
   */
  short               thik;
  /**
   *  Parameter pointer : Layer roughness 
   */
  short               sigz;
  /**
   *  Parameter pointer : Center of box/gaussian component 
   */
  short               cntr;
  /**
   *  Parameter pointer : Real value of the scattering length density
   */
  short               rsld;
  /**
   *  Parameter pointer : Imag. value of the scattering length density
   */
  short               isld;
  /**
   *  Parameter pointer : Magnetic scattering length density
   */
  short               rmag;
  /**
   *  Additional flags : Include the layer in the model?
   */
  enum boolean        include_flag;
  /**
   *  Additional flags : Used to distinguish repeating layers 
   *                     when writing output
   */
  unsigned short      repeat_flag;
  /**
   *  Additional flags : Polarization type. See \c polarized_type.
   */
  enum polarized_type polarized;
  /**
   *  Linked list : self-explained
   */
  struct _layer      *prev;
  /**
   *  Linked list : self-explained
   */
  struct _layer      *next;
};
typedef struct _layer yanera_layer; /**< See documentation for \p _layer*/
/** ---------------------------------------------------------------------------
 *  \brief The struct containing slab information.
 *
 *  Each model is deconstucted into tiny slabs by calling \c yanera_quadrature.
 *  Thus this struct contains only the information regarding slabs.
 *
 *  The maximum number of slabs is set to be the size of unsigned short maximum.
 */
struct _slabs
{
  unsigned short     number_of_slab_edges;  /**< Self explained */
  unsigned short     number_of_slabs;       /**< Self explained */
  /** Slab edge along from the postive z, ie. the slab end point */
  double             zpos[USHRT_MAX];
  /** Thickness of slab (slabs 0 and n-1 are inf.) */
  double             thik[USHRT_MAX];
  /** SLD of the slab */
  double             rsld[USHRT_MAX];
  /** ISLD of the slab */
  double             isld[USHRT_MAX];
};
typedef struct _slabs yanera_slabs; /**< See documentation for \p _slabs*/
/** ---------------------------------------------------------------------------
 *  \brief The struct for resolution information.
 *
 *  The struct for resolution information.
 */
struct _resolution {
  enum resolution_type  type;      /**< Type of instrumental resolution   */
  double               *value;     /**< Value of resolution               */
  double               *q;         /**< Array of q values for resolution  */
  short                 n;         /**< Number in array                   */
  char                 *file_name; /**< Filename for resolution array     */
};
/** See documentation for \p _resolution */
typedef struct _resolution yanera_resolution;
/** ---------------------------------------------------------------------------
 *  \brief The struct for data
 *
 * PLEASE FILL IN
 */
struct _data {
  /** Unique id linking dataset to a model. */
  char              *data_id;
  /** Filename containing the dataset. */
  char              *file_name;
  /** Number of datapoints */
  unsigned short     n;
  /** X-axis, or q, data points */
  double            *q;
  /** Y-axis, or R(q), data points */
  double            *R;
  /** Error in Y, or R(q), data  */
  double            *e;
  /** Index for fitting routines */
  short              idx;
  /** Instrumental solution associated with data. */
  yanera_resolution  resolution;
};
/**  See documentation for \p _data */
typedef struct _data yanera_data;
/** ---------------------------------------------------------------------------
 *   \brief The struct for each model 
 * 
 *
 *  This struct contains the linked list of layers that form the model,
 *  as well as additional information at the \<model/\> level, rather than
 *  \<layer/\> level.
 *
 *  Contains \c yanera_layer and \c yanera_slabs structs.
 */
struct _model {
  /**
   *  \<model/\> level detail : Unique id linking model to a dataset.
   */
  char               *data_idref;
  /**
   * \<model/\> level detail : Model name
   */
  char               *name;
  /**
   * \<model/\> level paramter : Background is a fitting parameter.
   * It's the only adjustable paramter at the model level.
   */
  short               background;
  /**
   * \<model/\> level detail : Polarization type. 
   */
  enum polarized_type polarized; 
  /**
   * Linked list of \<layer/\>
   */
  yanera_layer       *first_layer;
  /**
   * Linked list of \<layer/\>
   */
  yanera_layer       *last_layer;
  /**
   * Each model breaks down into small slabs
   */  
  yanera_slabs        slabs;
};
typedef struct _model yanera_model; /**< See documentation for \p _model*/
/** ---------------------------------------------------------------------------
 *  \brief All the adjustable parameters are in this struct.
 *
 *  All adjustable paramters, and whether they are fixed are held in 
 *  this struct.
 */
struct _parameters {
  double         *p;   /**< Array of all adjustable paramters */
  double         *min; /**< Constraint on parameters          */
  double         *max; /**< Constraint on parameters          */
  unsigned short *f;   /**< Array of flags constraints        */
  unsigned short *con; /**< Array of flags held fixed         */
  unsigned short  n;   /**< Number of paramters               */
  unsigned short  idx; /**< Index for fitting routines        */
  double         *tmp; /**< Temporary array of paramters      */
};
/** See documentation for \p _parameters */
typedef struct _parameters yanera_parameters;
/** ---------------------------------------------------------------------------
 *  \brief Miscellaneous data structure for data without catagory.
 *
 *  Miscellaneous data structure for data without catagory.
 */
struct _misc {
  /** Flag : Write out XML file of results */
  enum boolean             write_results;
  /** Flag : Write out to file SLD profiles (*.pro) */
  enum boolean             write_profile;
  /** Flag : Write out to file reflectivity curves (*.ref) */
  enum boolean             write_reflectivity;
  /** Flag : Write out to file SLD profiles slab approximations (*.slb)  */
  enum boolean             write_slabs;
  /** Flag : Slab creation tolerance See \c  _slabs for more information */
  double                   quadrature_error;
	/**  The max thickness of a quadrature slab is here */
  double                   quadrature_thik;
  /** Reflectivity curve output : Minimum \f$q\f$ value */
  double                   q_min;
  /** Reflectivity curve output : Maximum \f$q\f$ value  */
  double                   q_max;
  /** Reflectivity curve output  : Number of data points */
  short                    q_num;
  /** SLD profile curve output : Number of data points */
  short                    z_num;
  /** Flag : Preform fitting; if no, then just simulate */
  enum boolean                           fit_do;
  /** Fitting : Fit weighting */
  enum fit_weighting_type                fit_weighting;
  /** Fitting : Number of fitting iterations */
  short                                  fit_iterations;
  /** Fitting : Number of fitting iterations between updates */
  short                                  fit_write_iterations;
  /** Fitting : Fit weighting */
  enum fit_method_type                   fit_method;
  /** Fitting : Maximum number of iterations */
	const gsl_multimin_fdfminimizer_type  *fit_method_mdmin;
  /** Fitting : Maximum number of iterations */
	const gsl_multimin_fminimizer_type    *fit_method_mdmin_simplex;
  /** Fitting : Maximum number of iterations */
  const gsl_multifit_fdfsolver_type     *fit_method_lstsq;
#ifdef HAVE_OOL
  /** Fitting : Maximum number of iterations */
  const ool_conmin_minimizer_type       *fit_method_ool;
#endif
};
/**  See documentation for \p _misc */
typedef struct _misc yanera_misc;

/** ---------------------------------------------------------------------------
 *  \brief A container struct for holding everything.
 *
 *  This container holds everything. The idea is this: a pointer to a
 *  single \c yanera_container allocated in the main loop of the program 
 *  is passed down through the function calls.  Every function thus
 *  has access to all the stored information; but mostly it is 
 *  to allow access to the information in \c _misc and \c _parameters.
 *
 *  The data model followed here is closely related to the document model 
 *  of the XML input, that is to say, the first model is a complete 
 *  description of that model, and additional models only show modifications
 *  to that base model. Thus, the XML file is read into \c models_xml just as 
 *  they appear in the XML file. Then \c models_complete is contructed 
 *  from the \c models_xml. 
 *
 *  \bug There's no completeness checking of the complete models, although that 
 *       should be trivial.
 *
 *  Shared components are stored in \c component.
 *
 *  Access to paramters is simple, for example, to retrieve the real SLD 
 *  from a layer with either \c layer_id  "L" or \c name  "NAME" from within a 
 *  certain model with \c data_idref "M", one can try the psuedocode here
 *  
 *  \code
 *  yanera_container *y;
 *  yanera_layer     *l;
 *  \endcode
 *  First walk down the model list:
 *  \code
 *  for i=0 to y->number_of_models-1
 *  {
 *    if (y->model[i].idref == "M") break;
 *  }
 *  \endcode
 *  Now traverse the layer list:
 *  \code 
 *  l = y->model[i].layer;
 *  while ((l->id != "L") || (l->name != "NAME"))
 *  {
 *    l = l->next;
 *  }
 *  \endcode
 *  At this point, \c l now points to desired layer in desired model, and the 
 *  value of a parameter, \c rsld for example, can be easily obtained:
 *  \code
 *  y->parameters.p[l->rsld] 
 *  \endcode
 */
struct _yanera {
  /** 
   *  See \c model_type 
   */ 
  enum model_type    type;
  /** 
   *  PLEASE FILL IN 
   */
  double (*parrattFunction)(double, \
                            yanera_model *, \
                            struct _yanera *);
  /** 
   *  PLEASE FILL IN 
   */
  double (*profileFunctionReal)(double, \
                                yanera_model *, \
                                struct _yanera *);
  /** 
   *  PLEASE FILL IN 
   */
  double (*profileFunctionImag)(double, \
                                yanera_model *, \
                                struct _yanera *);
  /** 
   *  Linked list of models as specified by the XML input file
   *  See \c _model 
   */ 
  yanera_model       models_xml[10];
  /** 
   *  The complete models, one for every data set
   *  See \c _model 
   */
  yanera_model       models_complete[10];
  /** 
   *  The data sets to be fitted
   *  See \c _data
   */
  yanera_data        data[10];
  /**
   *  Self explanatory.
   */
  unsigned short     number_of_models;
  /**
   *  Self explanatory.
   */
  unsigned short     number_of_data;
  /**
   *  Index for fitting routines.
   */
  unsigned short     idx;
  /** 
   *  See \c _parameters 
   */
  yanera_parameters  parameters;
  /** 
   *  Linked list of common layer components.
   *  See \c _layer
   */
  yanera_layer      *components;
  /** 
   *  Miscellaneous data.
   *  See \c _misc
   */
  yanera_misc        misc;
  /**
   * XML filename
   */
  char              *filename;
};
/** See documentation for \p _yanera*/
typedef struct _yanera yanera_container;
/*====================================================================*/
/** \brief Set the Error Handler.

This function is called for terminating errors. Turn into a struct is you want to have non terminating error handling.
 */
void yanera_error(const char       *reason, \
                  const char       *file,   \
                  int               line,   \
                  yanera_container *yc);







