/*==============================================================================
Copyright 2006 Thad Harroun, Kevin Yager

This file is part of "yanera 2.0".

"yanera 2.0" is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version.

"yanera 2.0" is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with "yanera 2.0"; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
==============================================================================*/ 
/*============================================================================*/
/** \file yanera_profile.h
	  \brief Supplied functions for calculating the SLD profile.

This file has the functions required for calculating the SLD profile, \f$\rho(z)\f$.
 */
/*============================================================================*/
/** \brief Write the SLD profiles to file.

Write the SLD profiles, \f$\rho(z)\f$,  to file. The file names are drawn from the file names of the data files, and the file type suffix ".pro" is appended. If there were no data files, then the file name is "result_xx.pro", where xx is a sequential number based on the number of models.
 */
void yanera_write_profile(yanera_container *yanera);
/*============================================================================*/
/** \brief Calulates \f$\rho(z)\f$ for a slab model

  \param z Position \f$z\f$ to calculate \f$\rho(z)\f$
  \param model The \p yanera_model to calculate.
  \param yanera The general \p yanera_container.
  \return \f$\rho(z)\f$ 
	
For a set of \f$N\f$ layers, that run from bulk (\f$i=0\f$) to substrate (\f$i=N+1\f$) (or substrate to bulk), the scattering length density profile can be given as

\f{eqnarray*}
    \rho(z) &=& \rho_0 + \sum_{i=1}^{N+1}
              \frac{(\rho_i-\rho_{i-1})}{2}
              \left(
                 1+\mathrm{erf}\left(\frac{z-z_i}{\sqrt{2}\sigma_i}\right)
              \right)  \\
    z_i &=& \sum_{j=1}^{i} t_j
\f}

where \f$t_i\f$ is the thickness, \f$\rho_i\f$ is the real part of the scattering length density, \f$\sigma_i\f$ the roughness, each of layer \f$i\f$.  The location of the error function, \f$z_i\f$, is calculated by a running total of the thicknesses. When the roughness is zero, the error function becomes a Heavyside step function.
 */
double layerProfileReal(double z, yanera_model *model, \
                        yanera_container *yanera);
/*============================================================================*/
/** \brief Calulates \f$\rho(z)\f$ for a component model

  \param z Position \f$z\f$ to calculate \f$\rho(z)\f$
  \param model The \p yanera_model to calculate.
  \param yanera The general \p yanera_container.
  \return \f$\rho(z)\f$ 
  
For a set of \f$N\f$ layers, that run from bulk (\f$i=0\f$) to substrate (\f$i=N+1\f$) (or substrate to bulk), the scattering length density profile is a sum of the components that form it.

\f{displaymath}
    \rho(z) = \sum_{i=0}^{N+1} \rho_i \times F(z;\textrm{cntr},\textrm{sigz},\textrm{thik})
\f}

where \f$F(z)\f$ is a functional form of a component, that may have XML tags specifying the center (\p \<cntr/\>), roughness (\p \<sigz/\>), and thickness (\p \<thik/\>). 

The bulk and/or substrate are modeled by error functions centered at 0 and at the total thickness given by the \p \<cntr/\> tag for the bulk in the XML file.

\f{eqnarray*}
    \rho(z) &=& \textrm{rsld}_0 
              \left(
               1-\mathrm{erf}\left(\frac{z}{\sqrt{2}\textrm{sigz}_0}\right)
              \right)  \\
            &+& \textrm{rsld}_{N+1} 
              \left(
                1+\mathrm{erf}
  \left(\frac{z-\textrm{cntr}_{N+1}}{\sqrt{2}\textrm{sigz}_{N+1}}\right)
              \right)
\f}

For a Gaussian function, the \p \<sigz/\> tag becomes the width of the gaussian, given by
\f[
\textrm{rsld}\times e^{-(z-\textrm{cntr})/\textrm{sigz}^2} 
\f]

A box function is a box of total width \p \<thik/\>, convoluted with a Gaussian of width \p \<sigz/\>:
\f[
\textrm{rsld}\times\left[ 
\mathrm{erf}\left(\frac{z-(\textrm{cntr}-\textrm{thik}/2)}
                       {\sqrt{2}\cdot\textrm{sigz}}\right) -
\mathrm{erf}\left(\frac{z-(\textrm{cntr}+\textrm{thik}/2)}
                       {\sqrt{2}\cdot\textrm{sigz}}\right)                       
\right]
\f]
If \p \<sigz/\> is zero, then the box function becomes a Heavyside step function, and the Gaussian becomes zero.
 */
double componentProfileReal(double z, yanera_model *model, \
                            yanera_container *yanera);
/*============================================================================*/
/** \brief Calulates \f$\rho(z)\f$ for a function model

  \param z Position \f$z\f$ to calculate \f$\rho(z)\f$
  \param model The \p yanera_model to calculate.
  \param yanera The general \p yanera_container.
  \return \f$\rho(z)\f$ 
  
  This function calculates a  slab type profile model, as shown in  the function \c slabProfile, but for layers that have a functional form, that layer is evaluated according to the expression.
 */
double functionProfileReal(double z, yanera_model *model, \
                           yanera_container *yanera);
/*============================================================================*/
/** \brief Calulates imaginary component of \f$\rho(z)\f$ for a slab model

 \see slabProfileReal
 */
double layerProfileImag(double z, yanera_model *model, \
                        yanera_container *yanera);
/*============================================================================*/
/** \brief Calulates imaginary component of \f$\rho(z)\f$ for a component model

 \see componentProfileReal
 */
double componentProfileImag(double z, yanera_model *model, \
                            yanera_container *yanera);
/*============================================================================*/
/** \brief Calulates imaginary component of \f$\rho(z)\f$ for a function model

 \see functionProfileReal
 */
double functionProfileImag(double z, yanera_model *model, \
                           yanera_container *yanera);
/*============================================================================*/
/** \brief Figures the minimum and maximum \f$z\f$

  \param min Allocated double to contain the calculated minimum \f$z\f$
  \param max Allocated double to contain the calculated maximum \f$z\f$
  \param model The \p yanera_model to calculate.
  \param yanera The general \p yanera_container.

  Figures the minimum and maximum \f$z\f$ for the SLD profile. This is model ype dependant.
  \bug My confidence that this function works for all cases is about 95\%.
 */
void minMaxProfile(double *min, double *max, yanera_model *model, \
                   yanera_container *yanera);

