/*==============================================================================
Copyright 2006 Thad Harroun, Kevin Yager

This file is part of "yanera 2.0".

"yanera 2.0" is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version.

"yanera 2.0" is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with "yanera 2.0"; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 
==============================================================================*/ 
/*============================================================================*/
/** \file yanera_reflectivity.h
	  \brief Functions for calculating the reflectivity.

This file contains the declarations for the functions called for calculating the reflectivity curve. The formulas are defined in the reference page << \ref refl >>.
 */
/*============================================================================*/
/** \brief General function call to write refelctivity curves to file.

\param yanera The \c yanera_container.

General function call to write refelctivity curves to file. The function cycles through the models, creates an output filename for each, then Calculates and writes to file (on-the-fly) the reflectivity curve. The curve is calculated from \c q_min to \c q_max in \c q_num number of steps.

Filenames are based on the datafile names, with *.ref appended, or in the event of no datafiles, the default filename is "results_xx.ref" where xx is a sequential number from 1, based on how many models there are.
 */
void yanera_write_reflectivity(yanera_container *yanera);
/*============================================================================*/
/** \brief Wrapper for calculating the reflectivity with background and  instrumental resolution.

\param q The \f$q\f$ value to evaluate \f$R(q)\f$.
\param resolution A pointer to the \c yanera_resolution data for the datafile and the model.
\param model The \c yanera_model for which the refelctiity curve is calculated. 
\param yanera The \c yanera_container.
\returns The reflectivity \f$R(q)\f$. 

This function is a wrapper to add the background and instrument resolution. The actual reflectivity is calculated by calls to the function \p calcParratt() .

The instrumental resolution is determined by calculating the reflectivity averaged over several points with Gaussian weights. It's a method that was borrowed and made more general from \e Parratt32. It works by the following, assuming a \f$n\f$ point average,

\f{eqnarray*}
	R(q) &=& {\sum_{a=-n}^{n} w_a R(q[1+\frac{a}{n}\frac{\delta q}{q}])}/
	       {\sum_{a=-n}^{n} w_a} \\
	\mathrm{or} &=& {\sum_{a=-n}^{n} w_a R(q+\frac{a}{n}\delta q)}/
	       {\sum_{a=-n}^{n} w_a} \\
	w_a &=& \exp\left(-2\left(\frac{a}{n}\right)^2\right)
\f}

In the first equation, we can specify the instrumental resolution present in the data as in the same manner as it is typically recorded a Chalk River, \f$\delta q/q=\mathrm{const}\f$. Otherwise, the resolution can be specified as \f$\delta q=\mathrm{const}\f$. The Gaussian averaging of points has been pre-calculated for 9 point average.  If no resolution smearing is required, the calculation is skipped for speed.

Finally, the background is added.
 */
double calcReflectivity(double q, yanera_resolution *resolution, \
                        yanera_model *model, yanera_container *yanera);
/*============================================================================*/
/** \brief Calculates reflectivity \f$R(q)\f$ at \f$q\f$ from the slabs given in \c yanera_model

\param q The \f$q\f$ value to evaluate \f$R(q)\f$.
\param model The \c yanera_model for which the refelctiity curve is calculated. 
\param yanera The \c yanera_container.
\returns The reflectivity \f$R(q)\f$. 

Calculates reflectivity \f$R(q)\f$ at \f$q\f$ for given \c yanera_model. Formulas are located on page <<\ref refl>>. This function uses the tiny slabs calculated by adaptive quadrature.

NOTE: This function sets the imaginary SLD to zero.

At this point the imaginary part of the SLD is not being used. It is read in the XML but not calculated as part of the quadrature into little slabs, so is not accessible for use here. 
 */
double calcParrattFromSlabs(double q, \
                            yanera_model *model, \
                            yanera_container *yanera);
/*============================================================================*/
/** \brief Calculates reflectivity \f$R(q)\f$ at \f$q\f$ for given \c yanera_model

\param q The \f$q\f$ value to evaluate \f$R(q)\f$.
\param model The \c yanera_model for which the refelctiity curve is calculated. 
\param yanera The \c yanera_container.
\returns The reflectivity \f$R(q)\f$. 

Calculates reflectivity \f$R(q)\f$ at \f$q\f$ for given \c yanera_model. Formulas are located on page <<\ref refl>>. This functions calculates \f$R(q)\f$ directly from teh model, and not from the quadarture slabs.

NOTE: This function sets the imaginary SLD to zero.

At this point the imaginary part of the SLD is not being used. It is read in the XML but not calculated as part of the quadrature into little slabs, so is not accessible for use here. 
 */
double calcParrattFromModel(double q, \
                            yanera_model *model, \
                            yanera_container *yanera);
/*============================================================================*/
/** \brief Retruns the resolution spread at a point \f$q\f$. 

\param q The value of \f$q\f$ that we want the resolution
\param resolution A pointer to a \p yanera_rsolution structure with the resolution information.
\returns The resolution at that point.

Retruns the resolution spread at a point \f$q\f$.
 */
double lookupResolution(double  q, yanera_resolution *resolution);

